r"""Implement an analyzer that generates a section to analyze the number
of null values for each column."""

from __future__ import annotations

__all__ = ["ColumnTemporalNullValueAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnTemporalNullValueSection, EmptySection

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl


logger = logging.getLogger(__name__)


class ColumnTemporalNullValueAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the temporal distribution of null
    values for all columns.

    A plot is generated for each column.

    Args:
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        columns: The list of columns to analyze. A plot is generated
            for each column. ``None`` means all the columns.
        ncols: The number of columns.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.analyzer import ColumnTemporalNullValueAnalyzer
    >>> analyzer = ColumnTemporalNullValueAnalyzer("datetime", period="M")
    >>> analyzer
    ColumnTemporalNullValueAnalyzer(
      (columns): ()
      (dt_column): datetime
      (period): M
      (ncols): 2
      (figsize): (7, 5)
    )
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={
    ...         "int": pl.Int64,
    ...         "float": pl.Float64,
    ...         "str": pl.String,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    ColumnTemporalNullValueSection(
      (columns): ('float', 'int', 'str')
      (dt_column): datetime
      (period): M
      (ncols): 2
      (figsize): (7, 5)
    )

    ```
    """

    def __init__(
        self,
        dt_column: str,
        period: str,
        columns: Sequence[str] | None = None,
        ncols: int = 2,
        figsize: tuple[float, float] = (7, 5),
    ) -> None:
        self._dt_column = dt_column
        self._period = period
        self._columns = tuple(columns or [])
        self._ncols = ncols
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "columns": self._columns,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "ncols": self._ncols,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pl.DataFrame) -> ColumnTemporalNullValueSection | EmptySection:
        logger.info(
            "Analyzing the temporal null value distribution of all columns | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        if self._dt_column not in frame:
            logger.info(
                "Skipping monthly null value analysis because the datetime column "
                f"({self._dt_column}) is not in the DataFrame: {sorted(frame.columns)}"
            )
            return EmptySection()
        columns = self._columns if self._columns else sorted(frame.columns)
        if self._dt_column in columns:
            # Exclude the datetime column because it does not make sense to analyze it because
            # we cannot know the date/time if the value is null.
            columns.remove(self._dt_column)
        return ColumnTemporalNullValueSection(
            frame=frame,
            columns=columns,
            dt_column=self._dt_column,
            period=self._period,
            ncols=self._ncols,
            figsize=self._figsize,
        )



r"""Implement continuous values analyzers."""

from __future__ import annotations

__all__ = ["ColumnContinuousTemporalDriftAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnContinuousTemporalDriftSection, EmptySection

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class ColumnContinuousTemporalDriftAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the temporal drift of a column with
    continuous values.

    Args:
        column: The column name.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        nbins: The number of bins in the histogram.
        density: If True, draw and return a probability density:
            each bin will display the bin's raw count divided by the
            total number of counts and the bin width, so that the area
            under the histogram integrates to 1.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import numpy as np
    >>> import polars as pl
    >>> from flamme.analyzer import ColumnContinuousTemporalDriftAnalyzer
    >>> analyzer = ColumnContinuousTemporalDriftAnalyzer(
    ...     column="col", dt_column="date", period="1mo"
    ... )
    >>> analyzer
    ColumnContinuousTemporalDriftAnalyzer(
      (column): col
      (dt_column): date
      (period): 1mo
      (nbins): None
      (density): False
      (yscale): auto
      (xmin): None
      (xmax): None
      (figsize): None
    )
    >>> rng = np.random.default_rng()
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col": rng.standard_normal(59),
    ...         "datetime": pl.datetime_range(
    ...             start=datetime(year=2018, month=1, day=1, tzinfo=timezone.utc),
    ...             end=datetime(year=2018, month=3, day=1, tzinfo=timezone.utc),
    ...             interval="1d",
    ...             closed="left",
    ...             eager=True,
    ...         ),
    ...     },
    ...     schema={"col": pl.Int64, "datetime": pl.Datetime(time_unit="us", time_zone="UTC")},
    ... )
    >>> section = analyzer.analyze(frame)

    ```
    """

    def __init__(
        self,
        column: str,
        dt_column: str,
        period: str,
        nbins: int | None = None,
        density: bool = False,
        yscale: str = "auto",
        xmin: float | str | None = None,
        xmax: float | str | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._nbins = nbins
        self._density = density
        self._yscale = yscale
        self._xmin = xmin
        self._xmax = xmax
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "nbins": self._nbins,
                    "density": self._density,
                    "yscale": self._yscale,
                    "xmin": self._xmin,
                    "xmax": self._xmax,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pl.DataFrame) -> ColumnContinuousTemporalDriftSection | EmptySection:
        logger.info(f"Analyzing the temporal drift of {self._column}")
        for column in [self._column, self._dt_column]:
            if column not in frame:
                logger.info(
                    f"Skipping temporal drift analysis because the column ({column}) is not "
                    f"in the DataFrame"
                )
                return EmptySection()
        if self._column == self._dt_column:
            logger.info(
                "Skipping temporal continuous distribution analysis because the datetime column "
                f"({self._column}) is the column to analyze"
            )
            return EmptySection()
        return ColumnContinuousTemporalDriftSection(
            frame=frame,
            column=self._column,
            dt_column=self._dt_column,
            period=self._period,
            nbins=self._nbins,
            yscale=self._yscale,
            figsize=self._figsize,
            xmin=self._xmin,
            xmax=self._xmax,
            density=self._density,
        )



r"""Contain the implementation of a section to analyze a column with
discrete values."""

from __future__ import annotations

__all__ = [
    "ColumnDiscreteSection",
    "create_histogram",
    "create_histogram_section",
    "create_section_template",
    "create_table",
]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot import bar_discrete
from flamme.section.base import BaseSection
from flamme.section.most_frequent import create_frequent_values_table
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import MISSING_FIGURE_MESSAGE, figure2html

if TYPE_CHECKING:
    from collections import Counter
    from collections.abc import Sequence

    import polars as pl

logger = logging.getLogger(__name__)


class ColumnDiscreteSection(BaseSection):
    r"""Implement a section that analyzes a discrete distribution of
    values.

    Args:
        counter: The counter that represents the discrete
            distribution.
        null_values: The number of null values.
        dtype: The column data type.
        column: The column name.
        max_rows: The maximum number of rows to show in the
            table.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> from flamme.section import ColumnDiscreteSection
    >>> section = ColumnDiscreteSection(counter=Counter({"a": 4, "b": 2, "c": 6}), column="col")
    >>> section
    ColumnDiscreteSection(
      (null_values): 0
      (column): col
      (yscale): auto
      (max_rows): 20
      (figsize): None
    )
    >>> section.get_statistics()
    {'most_common': [('c', 6), ('a', 4), ('b', 2)], 'null_values': 0, 'nunique': 3, 'total': 12}

    ```
    """

    def __init__(
        self,
        counter: Counter,
        null_values: int = 0,
        dtype: pl.DataType | None = None,
        column: str = "N/A",
        max_rows: int = 20,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._counter = counter
        self._null_values = null_values
        self._column = column
        self._max_rows = int(max_rows)
        self._yscale = yscale
        self._figsize = figsize
        self._dtype = dtype

        self._total = sum(self._counter.values())

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "null_values": self._null_values,
                    "column": self._column,
                    "yscale": self._yscale,
                    "max_rows": self._max_rows,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    @property
    def yscale(self) -> str:
        return self._yscale

    def get_statistics(self) -> dict:
        most_common = [(value, count) for value, count in self._counter.most_common() if count > 0]
        return {
            "most_common": most_common,
            "null_values": self._null_values,
            "nunique": len(most_common),
            "total": self._total,
        }

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the discrete distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * self._null_values / stats['total']:.2f}" if stats["total"] > 0 else "N/A"
        )
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "total_values": f"{stats['total']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{self._null_values:,}",
                "null_values_pct": null_values_pct,
                "figure": create_histogram_section(
                    counter=self._counter,
                    column=self._column,
                    yscale=self._yscale,
                    figsize=self._figsize,
                ),
                "table": create_table(
                    column=self._column,
                    counter=self._counter,
                    max_rows=self._max_rows,
                ),
                "dtype": str(self._dtype),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.discrete import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the discrete distribution of values for column <em>{{column}}</em>.

<ul>
<li> <b>total values:</b> {{total_values}} </li>
<li> <b>number of unique values:</b> {{unique_values}} </li>
<li> <b>number of null values:</b> {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
<li> <b>data type:</b> <em>{{dtype}}</em> </li>
</ul>

{{figure}}
{{table}}

<p style="margin-top: 1rem;">
"""


def create_histogram_section(
    counter: Counter,
    column: str = "N/A",
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> str:
    r"""Return the histogram section.

    Args:
        counter: The counter that represents the discrete
            distribution.
        column: The column name.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The histogram section.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> from flamme.section.discrete import create_histogram_section
    >>> section = create_histogram_section(
    ...     counter=Counter({"a": 4, "b": 2, "c": 6}), column="col"
    ... )

    ```
    """
    if sum(counter.values()) == 0:
        return MISSING_FIGURE_MESSAGE
    most_common = [
        (value, count) for value, count in counter.most_common() if count > 0 and value is not None
    ]
    fig = create_histogram(
        column=column,
        names=[str(value) for value, _ in most_common],
        counts=[count for _, count in most_common],
        yscale=yscale,
        figsize=figsize,
    )
    return Template(
        r"""<p style="margin-top: 1rem;">
<b>Distribution of values in column {{column}}</b>

<p>The values in the figure below are sorted by decreasing order of number of occurrences.

{{figure}}
"""
    ).render({"figure": figure2html(fig, close_fig=True), "column": column})


def create_histogram(
    column: str,
    names: Sequence,
    counts: Sequence[int],
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Return a figure with the histogram of discrete values.

    Args:
        column: The column name.
        names: The name of the values to plot.
        counts: The number of value occurrences.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure.

    Example usage:

    ```pycon

    >>> from flamme.section.discrete import create_histogram
    >>> fig = create_histogram(
    ...     column="col", names=["a", "b", "c", "d"], counts=[5, 100, 42, 27]
    ... )

    ```
    """
    if sum(counts) == 0:
        return None
    fig, ax = plt.subplots(figsize=figsize)
    bar_discrete(ax=ax, names=names, counts=counts, yscale=yscale)
    ax.set_title(f"number of occurrences for each value of {column}")
    return fig


def create_table(
    counter: Counter,
    column: str = "N/A",
    max_rows: int = 20,
) -> str:
    r"""Return a HTML table with the discrete distribution.

    Args:
        counter: The counter that represents the discrete
            distribution.
        column: The column name.
        max_rows: The maximum number of rows to show in the
            table.

    Returns:
        The generated table.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> from flamme.section.discrete import create_table
    >>> table = create_table(counter=Counter({"a": 4, "b": 2, "c": 6}), column="col")

    ```
    """
    if sum(counter.values()) == 0:
        return "<span>&#9888;</span> No table is generated because the column is empty"

    return Template(
        """<details>
    <summary>[show head and tail values]</summary>

    <ul>
      <li> <b>count</b>: is the number of occurrences of the value </li>
      <li> <b>percentage</b>: is the number of occurrences the value divided by the total number of occurrences </li>
      <li> <b>cumulative percentage</b>: is the sum of the percentage of occurrence for the value plus all the previous values in the table. </li>
    </ul>

    <div class="row">
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Head: {{max_values}} most common values in column <em>{{column}}</em></b>
        {{table_head}}
      </div>
      <div class="col">
        <p style="margin-top: 1rem;">
        <b>Tail: {{max_values}} least common values in column <em>{{column}}</em></b>
        {{table_tail}}
      </div>
    </div>
</details>
"""
    ).render(
        {
            "max_values": len(counter.most_common(max_rows)),
            "table_head": create_frequent_values_table(counter=counter, top=max_rows),
            "table_tail": create_frequent_values_table(counter=counter, top=max_rows, reverse=True),
            "column": column,
        }
    )



r"""Contain the implementation of a section to count the number of rows
for a given temporal window."""

from __future__ import annotations

__all__ = [
    "TemporalRowCountSection",
    "create_section_template",
    "create_temporal_count_figure",
    "create_temporal_count_table",
    "create_temporal_count_table_row",
]

import logging
from typing import TYPE_CHECKING

from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot.utils import readable_xticklabels
from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.count import compute_temporal_count
from flamme.utils.figure import figure2html

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl


logger = logging.getLogger(__name__)


class TemporalRowCountSection(BaseSection):
    r"""Implement a section to analyze the number of rows per temporal
    window.

    Args:
        frame: The DataFrame to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section import TemporalRowCountSection
    >>> section = TemporalRowCountSection(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ]
    ...         },
    ...         schema={"datetime": pl.Datetime(time_unit="us", time_zone="UTC")},
    ...     ),
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )
    >>> section
    TemporalRowCountSection(dt_column=datetime, period=1mo, figsize=None)
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        if dt_column not in frame:
            msg = (
                f"Datetime column {dt_column} is not in the DataFrame "
                f"(columns:{sorted(frame.columns)})"
            )
            raise ValueError(msg)
        self._frame = frame
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(dt_column={self._dt_column}, period={self._period}, "
            f"figsize={self._figsize})"
        )

    @property
    def frame(self) -> pl.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def dt_column(self) -> str:
        r"""The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""The temporal period used to analyze the data."""
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            "Rendering the number of rows per temporal window "
            f"| datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "dt_column": self._dt_column,
                "figure": self._create_temporal_count_figure(),
                "table": create_temporal_count_table(
                    frame=self._frame,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_temporal_count_figure(self) -> str:
        fig = create_temporal_count_figure(
            frame=self._frame, dt_column=self._dt_column, period=self._period, figsize=self._figsize
        )
        return figure2html(fig, close_fig=True)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.count_rows import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the number of rows per temporal window.
The column <em>{{dt_column}}</em> is used as the temporal column.

{{figure}}

{{table}}

<p style="margin-top: 1rem;">
"""


def create_temporal_count_figure(
    frame: pl.DataFrame,
    dt_column: str,
    period: str,
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Return a figure with number of rows per temporal windows.

    Args:
        frame: The DataFrame to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.count_rows import create_temporal_count_figure
    >>> fig = create_temporal_count_figure(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ]
    ...         },
    ...         schema={"datetime": pl.Datetime(time_unit="us", time_zone="UTC")},
    ...     ),
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )

    ```
    """
    if frame.is_empty() or dt_column not in frame:
        return None

    counts, labels = compute_temporal_count(frame=frame, dt_column=dt_column, period=period)
    fig, ax = plt.subplots(figsize=figsize)
    ax.bar(x=labels, height=counts, color="tab:blue")
    ax.set_ylabel("number of rows")
    ax.set_xlim(-0.5, len(labels) - 0.5)
    readable_xticklabels(ax, max_num_xticks=100)
    return fig


def create_temporal_count_table(frame: pl.DataFrame, dt_column: str, period: str) -> str:
    r"""Return a HTML representation of a figure with number of rows per
    temporal windows.

    Args:
        frame: The DataFrame to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.

    Returns:
        The HTML representation of the table.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.count_rows import create_temporal_count_table
    >>> table = create_temporal_count_table(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ]
    ...         },
    ...         schema={"datetime": pl.Datetime(time_unit="us", time_zone="UTC")},
    ...     ),
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )

    ```
    """
    if frame.is_empty():
        return ""
    counts, labels = compute_temporal_count(frame=frame, dt_column=dt_column, period=period)
    rows = [
        create_temporal_count_table_row(label=label, num_rows=num_rows)
        for label, num_rows in zip(labels, counts)
    ]
    return Template(
        """<details>
    <summary>[show statistics per temporal period]</summary>

    <p>The following table shows some statistics for each period.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>period</th>
                <th>number of rows</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "period": period})


def create_temporal_count_table_row(label: str, num_rows: int) -> str:
    r"""Return the HTML code of a table row.

    Args:
        label: The label i.e. temporal window.
        num_rows: The number of rows for the given temporal
            window.

    Returns:
        The HTML code of a row.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.count_rows import create_temporal_count_table_row
    >>> row = create_temporal_count_table_row(label="meow", num_rows=42)

    ```
    """
    return Template("<tr><th>{{label}}</th><td {{num_style}}>{{num_rows}}</td></tr>").render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "num_rows": f"{num_rows:,}",
        }
    )



r"""Contain plotting functions to analyze continuous values."""

from __future__ import annotations

__all__ = [
    "boxplot_continuous",
    "boxplot_continuous_temporal",
    "hist_continuous",
    "hist_continuous2",
]

from typing import TYPE_CHECKING

import numpy as np

from flamme.plot.cdf import plot_cdf
from flamme.plot.utils import (
    auto_yscale_continuous,
    axvline_quantile,
    readable_xticklabels,
)
from flamme.utils.array import nonnan
from flamme.utils.range import find_range

if TYPE_CHECKING:
    from collections.abc import Sequence

    from matplotlib.axes import Axes


def boxplot_continuous(
    ax: Axes,
    array: np.ndarray,
    xmin: float | str | None = None,
    xmax: float | str | None = None,
) -> None:
    r"""Plot the histogram of an array containing continuous values.

    Args:
        ax: The axes of the matplotlib figure to update.
        array: The array with the data.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import boxplot_continuous
    >>> fig, ax = plt.subplots()
    >>> boxplot_continuous(ax, array=np.arange(101))

    ```
    """
    array = array.ravel()
    if array.size == 0:
        return
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    ax.boxplot(
        array,
        notch=True,
        vert=False,
        widths=0.7,
        patch_artist=True,
        boxprops={"facecolor": "lightblue"},
    )
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_ylabel(" ")


def boxplot_continuous_temporal(
    ax: Axes,
    data: Sequence[np.ndarray],
    steps: Sequence,
    ymin: float | str | None = None,
    ymax: float | str | None = None,
    yscale: str = "linear",
) -> None:
    r"""Plot the histogram of an array containing continuous values.

    Args:
        ax: The axes of the matplotlib figure to update.
        data: The sequence of data where each item is a 1-d array with
            the values of the time step.
        steps: The sequence time step names.
        ymin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        ymax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.

    Raises:
        RuntimeError: if ``data`` and ``steps`` have different lengths

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import boxplot_continuous_temporal
    >>> fig, ax = plt.subplots()
    >>> rng = np.random.default_rng()
    >>> data = [rng.standard_normal(1000) for _ in range(10)]
    >>> boxplot_continuous_temporal(ax, data=data, steps=list(range(len(data))))

    ```
    """
    if len(data) == 0:
        return
    if len(data) != len(steps):
        msg = f"data and steps have different lengths: {len(data):,} vs {len(steps):,}"
        raise RuntimeError(msg)
    data = [nonnan(x) for x in data]
    ax.boxplot(
        data,
        notch=True,
        vert=True,
        widths=0.7,
        patch_artist=True,
        boxprops={"facecolor": "lightblue"},
    )
    array = np.concatenate(data)
    ymin, ymax = find_range(array, xmin=ymin, xmax=ymax)
    if ymin < ymax:
        ax.set_ylim(ymin, ymax)
    ax.set_xticks(np.arange(len(steps)), labels=steps)
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=100)
    ax.set_yscale(yscale)
    readable_xticklabels(ax)


def hist_continuous(
    ax: Axes,
    array: np.ndarray,
    nbins: int | None = None,
    density: bool = False,
    yscale: str = "linear",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    cdf: bool = True,
    quantile: bool = True,
) -> None:
    r"""Plot the histogram of an array containing continuous values.

    Args:
        ax: The axes of the matplotlib figure to update.
        array: The array with the data.
        nbins: The number of bins to use to plot.
        density: If True, draw and return a probability density:
            each bin will display the bin's raw count divided by the
            total number of counts and the bin width, so that the area
            under the histogram integrates to 1.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        cdf: If ``True``, the CDF is added to the plot.
        quantile: If ``True``, the 5% and 95% quantiles are added to
            the plot.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import hist_continuous
    >>> fig, ax = plt.subplots()
    >>> hist_continuous(ax, array=np.arange(101))

    ```
    """
    array = array.ravel()
    if array.size == 0:
        return
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    ax.hist(array, bins=nbins, range=(xmin, xmax), color="tab:blue", alpha=0.9, density=density)
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_ylabel("density (number of occurrences/total)" if density else "number of occurrences")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=nbins)
    ax.set_yscale(yscale)
    if cdf:
        plot_cdf(
            ax=ax.twinx(),
            array=array,
            nbins=nbins,
            xmin=xmin,
            xmax=xmax,
            color="tab:red",
            labelcolor="tab:red",
        )

    if not quantile:
        return
    q05, q95 = np.quantile(array, q=[0.05, 0.95])
    if xmin < q05 < xmax:
        axvline_quantile(ax, quantile=q05, label="q0.05 ", horizontalalignment="right")
    if xmin < q95 < xmax:
        axvline_quantile(ax, quantile=q95, label=" q0.95", horizontalalignment="left")


def hist_continuous2(
    ax: Axes,
    array1: np.ndarray,
    array2: np.ndarray,
    label1: str = "first",
    label2: str = "second",
    nbins: int | None = None,
    density: bool = False,
    yscale: str = "linear",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
) -> None:
    r"""Plot the histogram of two arrays to compare the distributions.

    Args:
        ax: The axes of the matplotlib figure to update.
        array1: The first array with the data.
        array2: The second array with the data.
        label1: The label associated to the first array.
        label2: The label associated to the second array.
        nbins: The number of bins to use to plot.
        density: If True, draw and return a probability density:
            each bin will display the bin's raw count divided by the
            total number of counts and the bin width, so that the area
            under the histogram integrates to 1.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import hist_continuous2
    >>> fig, ax = plt.subplots()
    >>> hist_continuous2(ax, array1=np.arange(101), array2=np.arange(51))

    ```
    """
    array1, array2 = array1.ravel(), array2.ravel()
    array = np.concatenate([array1, array2])
    if array.size == 0:
        return
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    ax.hist(
        array1,
        bins=nbins,
        range=(xmin, xmax),
        color="tab:blue",
        alpha=0.5,
        label=label1,
        density=density,
    )
    ax.hist(
        array2,
        bins=nbins,
        range=(xmin, xmax),
        color="tab:orange",
        alpha=0.5,
        label=label2,
        density=density,
    )
    readable_xticklabels(ax, max_num_xticks=100)
    if xmin < xmax:
        ax.set_xlim(xmin, xmax)
    ax.set_ylabel("density (number of occurrences/total)" if density else "number of occurrences")
    if yscale == "auto":
        yscale = auto_yscale_continuous(array=array, nbins=nbins)
    ax.set_yscale(yscale)
    ax.legend()



r"""Implement an analyzer that generates a markdown section."""

from __future__ import annotations

__all__ = ["MarkdownAnalyzer"]

from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import MarkdownSection
from flamme.utils.imports import check_markdown

if TYPE_CHECKING:
    import polars as pl


class MarkdownAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that adds a mardown string to the report.

    Args:
        desc: The markdown description.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.analyzer import MarkdownAnalyzer
    >>> analyzer = MarkdownAnalyzer(desc="hello cats!")
    >>> analyzer
    MarkdownAnalyzer()
    >>> frame = pl.DataFrame(
    ...     {
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "int": [None, 1, 0, 1],
    ...         "str": ["A", "B", None, None],
    ...     },
    ...     schema={"float": pl.Float64, "int": pl.Int64, "str": pl.String},
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    MarkdownSection()

    ```
    """

    def __init__(self, desc: str) -> None:
        check_markdown()
        self._desc = str(desc)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, frame: pl.DataFrame) -> MarkdownSection:  # noqa: ARG002
        return MarkdownSection(desc=self._desc)



r"""Contain CDF plotting functions."""

from __future__ import annotations

__all__ = ["plot_cdf"]

from typing import TYPE_CHECKING

import numpy as np

from flamme.utils.array import nonnan

if TYPE_CHECKING:
    from matplotlib.axes import Axes


def plot_cdf(
    ax: Axes,
    array: np.ndarray,
    nbins: int | None = None,
    xmin: float = float("-inf"),
    xmax: float = float("inf"),
    color: str = "tab:blue",
    labelcolor: str = "black",
) -> None:
    r"""Plot the cumulative distribution function (CDF).

    Args:
        ax: The axes of the matplotlib figure to update.
        array: The array with the data.
        nbins: The number of bins to use to plot the CDF.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        color: The plot color.
        labelcolor: The label color.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import plot_cdf
    >>> fig, ax = plt.subplots()
    >>> plot_cdf(ax, array=np.arange(101))

    ```
    """
    array = nonnan(array.ravel())
    if array.size == 0:
        return
    nbins = nbins or min(1000, array.size)
    nleft = array[array < xmin].size
    nright = array[array > xmax].size
    counts, edges = np.histogram(array[np.logical_and(array >= xmin, array <= xmax)], bins=nbins)
    cdf = (np.cumsum(counts) + nleft) / (np.sum(counts) + nleft + nright)
    x = [(left + right) * 0.5 for left, right in zip(edges[:-1], edges[1:])]
    ax.tick_params(axis="y", labelcolor=labelcolor)
    ax.plot(x, cdf, color=color, label="CDF")
    ax.set_ylim(0.0, 1.0)
    ax.set_ylabel("cumulative distribution function (CDF)", color=labelcolor)



r"""Implement an analyzer that transforms the data before to analyze the
data."""

from __future__ import annotations

__all__ = ["TransformAnalyzer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import str_indent, str_mapping
from grizz.transformer import BaseTransformer, setup_transformer

from flamme.analyzer.base import BaseAnalyzer, setup_analyzer

if TYPE_CHECKING:
    import polars as pl

    from flamme.section import BaseSection

logger = logging.getLogger(__name__)


class TransformAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer that filters the data before to analyze the
    data.

    Args:
        transformer: Soecifies the transformer.
        analyzer: The analyzer or its configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.analyzer import TransformAnalyzer, NullValueAnalyzer
    >>> from grizz.transformer import SqlTransformer
    >>> analyzer = TransformAnalyzer(
    ...     transformer=SqlTransformer("SELECT * FROM self WHERE float > 1"),
    ...     analyzer=NullValueAnalyzer(),
    ... )
    >>> analyzer
    TransformAnalyzer(
      (transformer): SqlTransformer(
          (query): SELECT * FROM self WHERE float > 1
        )
      (analyzer): NullValueAnalyzer(figsize=None)
    )
    >>> frame = pl.DataFrame(
    ...     {
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "int": [None, 1, 0, 1],
    ...         "str": ["A", "B", None, None],
    ...     },
    ...     schema={"float": pl.Float64, "int": pl.Int64, "str": pl.String},
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    NullValueSection(
      (columns): ('float', 'int', 'str')
      (null_count): array([0, 1, 1])
      (total_count): array([3, 3, 3])
      (figsize): None
    )

    ```
    """

    def __init__(self, transformer: BaseTransformer | dict, analyzer: BaseAnalyzer | dict) -> None:
        self._transformer = setup_transformer(transformer)
        self._analyzer = setup_analyzer(analyzer)

    def __repr__(self) -> str:
        args = str_indent(
            str_mapping({"transformer": self._transformer, "analyzer": self._analyzer})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def analyze(self, frame: pl.DataFrame) -> BaseSection:
        logger.info("Transforming the DataFrame...")
        frame = self._transformer.transform(frame)
        return self._analyzer.analyze(frame)



r"""Contain the implementation of a section to analyze the most frequent
values for a given columns."""

from __future__ import annotations

__all__ = [
    "MostFrequentValuesSection",
    "create_section_template",
    "create_frequent_values_table",
    "create_table_row",
]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template

from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)

if TYPE_CHECKING:
    from collections import Counter
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class MostFrequentValuesSection(BaseSection):
    r"""Implement a section that analyzes the most frequent values for a
    given columns.

    Args:
        counter: The counter with the number of occurrences
            for all values.
        column: The column name.
        top: The maximum number of values to show.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> from flamme.section import MostFrequentValuesSection
    >>> section = MostFrequentValuesSection(
    ...     counter=Counter({"a": 4, "b": 2, "c": 6}), column="col"
    ... )
    >>> section
    MostFrequentValuesSection(
      (counter): Counter({'c': 6, 'a': 4, 'b': 2})
      (column): col
      (top): 100
      (total): 12
    )
    >>> section.get_statistics()
    {'most_common': [('c', 6), ('a', 4), ('b', 2)]}

    ```
    """

    def __init__(self, counter: Counter, column: str, top: int = 100) -> None:
        self._counter = counter
        self._column = column
        self._top = top

        self._total = sum(self._counter.values())

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "counter": self._counter,
                    "column": self._column,
                    "top": self._top,
                    "total": self._total,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def get_statistics(self) -> dict:
        return {"most_common": list(self._counter.most_common(self._top))}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info("Rendering the most frequent values section...")
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "table": create_frequent_values_table(counter=self._counter, top=self._top),
                "column": self._column,
                "top": f"{self._top:,}",
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.most_frequent import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the {{top}} most frequent values in <em>{{column}}</em>.

<ul>
  <li> <b>count</b>: is the number of occurrences of the value </li>
  <li> <b>percentage</b>: is the number of occurrences the value divided by the total number of occurrences </li>
  <li> <b>cumulative percentage</b>: is the sum of the percentage of occurrence for the value plus all the previous values in the table. </li>
</ul>

{{table}}

<p style="margin-top: 1rem;">
"""


def create_frequent_values_table(counter: Counter, top: int = 100, reverse: bool = False) -> str:
    r"""Return a HTML representation of a table with the most (or least)
    frequent values.

    Args:
        counter: The counter with the number of occurrences
            for all values.
        top: The maximum number of values to show.
        reverse: If ``True``, it returns a table with the least
            frequent values (a.k.a. tail), otherwise it returns a
            table with the most frequent values (a.k.a. head).

    Returns:
        The HTML representation of the table.

    Example usage:

    ```pycon

    >>> from collections import Counter
    >>> from flamme.section.most_frequent import create_frequent_values_table
    >>> table = create_frequent_values_table(Counter({"a": 4, "b": 2, "c": 6}))

    ```
    """
    total = sum(counter.values())
    most_common = counter.most_common()[-top:][::-1] if reverse else counter.most_common(top)
    rows = []
    cumcount = 0
    for value, count in most_common:
        cumcount += count
        rows.append(create_table_row(value=value, count=count, total=total, cumcount=cumcount))
    return Template(
        """<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th>value</th>
            <th>count</th>
            <th>percentage (%)</th>
            <th>cumulative percentage (%)</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
    ).render({"rows": "\n".join(rows)})


def create_table_row(value: str, count: int, total: int, cumcount: int) -> str:
    r"""Create the HTML code of a new table row.

    Args:
        value: Specifies a string representation of the value.
        count: The number of occurrences of the value.
        total: The total number of occurrences.
        cumcount: The cumulative number of occurrences.

    Returns:
        The HTML code of a row.

    Example usage:

    ```pycon

    >>> from flamme.section.most_frequent import create_table_row
    >>> row = create_table_row(value="A", count=5, total=101, cumcount=42)

    ```
    """
    pct = 100 * count / total if total > 0 else float("nan")
    cum_percentage = 100 * cumcount / total if total > 0 else float("nan")
    return Template(
        """<tr>
    <th>{{value}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{percentage}}</td>
    <td {{num_style}}>{{cum_percentage}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "value": value,
            "count": f"{count:,}",
            "percentage": f"{pct:.2f}",
            "cum_percentage": f"{cum_percentage:.2f}",
        }
    )



r"""Implement discrete values analyzers."""

from __future__ import annotations

__all__ = ["ColumnDiscreteAnalyzer"]

import logging
from collections import Counter
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnDiscreteSection, EmptySection

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class ColumnDiscreteAnalyzer(BaseAnalyzer):
    r"""Implement a discrete distribution analyzer.

    Args:
        column: The column to analyze.
        drop_nulls: If ``True``, the NaN values are not included in the
            analysis.
        max_rows: The maximum number of rows to show in the
            table.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.analyzer import ColumnDiscreteAnalyzer
    >>> analyzer = ColumnDiscreteAnalyzer(column="str")
    >>> analyzer
    ColumnDiscreteAnalyzer(column=str, drop_nulls=False, max_rows=20, yscale=auto, figsize=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...     }
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    ColumnDiscreteSection(
      (null_values): 2
      (column): str
      (yscale): auto
      (max_rows): 20
      (figsize): None
    )

    ```
    """

    def __init__(
        self,
        column: str,
        drop_nulls: bool = False,
        max_rows: int = 20,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._drop_nulls = bool(drop_nulls)
        self._max_rows = max_rows
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"drop_nulls={self._drop_nulls}, max_rows={self._max_rows}, yscale={self._yscale}, "
            f"figsize={self._figsize})"
        )

    def analyze(self, frame: pl.DataFrame) -> ColumnDiscreteSection | EmptySection:
        logger.info(f"Analyzing the discrete distribution of {self._column}")
        if self._column not in frame:
            logger.info(
                f"Skipping discrete distribution analysis of column {self._column} "
                f"because it is not in the DataFrame: {sorted(frame.columns)}"
            )
            return EmptySection()
        series = frame[self._column]
        if self._drop_nulls:
            series = series.drop_nulls()
        return ColumnDiscreteSection(
            counter=Counter(series.to_list()),
            null_values=series.null_count(),
            dtype=series.dtype,
            column=self._column,
            max_rows=self._max_rows,
            yscale=self._yscale,
            figsize=self._figsize,
        )



r"""Contain the implementation of a section to analyze the temporal
distribution of a column with continuous values."""

from __future__ import annotations

__all__ = [
    "ColumnTemporalContinuousSection",
    "create_section_template",
    "create_temporal_figure",
    "create_temporal_table",
    "create_temporal_table_row",
]

import logging
from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot import boxplot_continuous_temporal
from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import figure2html
from flamme.utils.temporal import compute_temporal_stats, to_step_names

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class ColumnTemporalContinuousSection(BaseSection):
    r"""Implement a section that analyzes the temporal distribution of a
    column with continuous values.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section import ColumnContinuousSection
    >>> section = ColumnTemporalContinuousSection(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
    ...             "datetime": [
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=1, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=2, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Float64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )
    >>> section
    ColumnTemporalContinuousSection(
      (column): col
      (dt_column): datetime
      (period): 1mo
      (yscale): auto
      (figsize): None
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        column: str,
        dt_column: str,
        period: str,
        yscale: str = "auto",
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._frame = frame
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._yscale = yscale
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "yscale": self._yscale,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def period(self) -> str:
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal continuous distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure": self._create_temporal_figure(),
                "table": create_temporal_table(
                    frame=self._frame,
                    column=self._column,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_temporal_figure(self) -> str:
        fig = create_temporal_figure(
            frame=self._frame,
            column=self._column,
            dt_column=self._dt_column,
            period=self._period,
            yscale=self._yscale,
            figsize=self._figsize,
        )
        return figure2html(fig, close_fig=True)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.continuous_temp import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column <em>{{column}}</em>
by using the column <em>{{dt_column}}</em>.

{{figure}}

{{table}}

<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    frame: pl.DataFrame,
    column: str,
    dt_column: str,
    period: str,
    yscale: str = "auto",
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Create a figure with the temporal value distribution.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.continuous_temp import create_temporal_figure
    >>> fig = create_temporal_figure(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
    ...             "datetime": [
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=1, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=2, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Float64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )

    ```
    """
    if frame.is_empty():
        return None
    groups = (
        frame.select(pl.col(dt_column).alias("datetime"), pl.col(column).alias("value"))
        .sort("datetime")
        .group_by_dynamic("datetime", every=period)
    )
    data = [np.sort(np.array(x["value"])) for _, x in groups]
    steps = to_step_names(groups=groups, period=period)
    fig, ax = plt.subplots(figsize=figsize)
    boxplot_continuous_temporal(ax=ax, data=data, steps=steps, yscale=yscale)
    return fig


def create_temporal_table(frame: pl.DataFrame, column: str, dt_column: str, period: str) -> str:
    r"""Return a HTML representation of a table with some statistics
    about the temporal value distribution.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        The HTML representation of the table.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.continuous_temp import create_temporal_table
    >>> table = create_temporal_table(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
    ...             "datetime": [
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=1, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=2, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Float64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )

    ```
    """
    if frame.is_empty():
        return "<span>&#9888;</span> No table is generated because the column is empty"

    stats = compute_temporal_stats(frame=frame, column=column, dt_column=dt_column, period=period)

    rows = [create_temporal_table_row(stat) for stat in stats.to_dicts()]
    return Template(
        """<details>
    <summary>[show statistics per temporal period]</summary>

    <p>The following table shows some statistics for each period of column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr>
                <th>step</th>
                <th>count</th>
                <th>mean</th>
                <th>std</th>
                <th>min</th>
                <th>quantile 1%</th>
                <th>quantile 5%</th>
                <th>quantile 10%</th>
                <th>quantile 25%</th>
                <th>median</th>
                <th>quantile 75%</th>
                <th>quantile 90%</th>
                <th>quantile 95%</th>
                <th>quantile 99%</th>
                <th>max</th>
            </tr>
        </thead>
        <tbody class="tbody table-group-divider">
            {{rows}}
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_table_row(stats: dict) -> str:
    r"""Return the HTML code of a new table row.

    Args:
        stats: The statistics for the row.

    Returns:
        The HTML code of a row.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.continuous_temp import create_temporal_table_row
    >>> row = create_temporal_table_row(
    ...     stats={
    ...         "step": "2020-01-01",
    ...         "count": 101,
    ...         "nunique": 101,
    ...         "mean": 50.0,
    ...         "std": 29.300170647967224,
    ...         "min": 0.0,
    ...         "q01": 1.0,
    ...         "q05": 5.0,
    ...         "q10": 10.0,
    ...         "q25": 25.0,
    ...         "median": 50.0,
    ...         "q75": 75.0,
    ...         "q90": 90.0,
    ...         "q95": 95.0,
    ...         "q99": 99.0,
    ...         "max": 100.0,
    ...     }
    ... )

    ```
    """

    def to_float(value: float | None) -> float:
        if value is None:
            return float("nan")
        return value

    return Template(
        """<tr>
    <th>{{step}}</th>
    <td {{num_style}}>{{count}}</td>
    <td {{num_style}}>{{mean}}</td>
    <td {{num_style}}>{{std}}</td>
    <td {{num_style}}>{{min}}</td>
    <td {{num_style}}>{{q01}}</td>
    <td {{num_style}}>{{q05}}</td>
    <td {{num_style}}>{{q10}}</td>
    <td {{num_style}}>{{q25}}</td>
    <td {{num_style}}>{{median}}</td>
    <td {{num_style}}>{{q75}}</td>
    <td {{num_style}}>{{q90}}</td>
    <td {{num_style}}>{{q95}}</td>
    <td {{num_style}}>{{q99}}</td>
    <td {{num_style}}>{{max}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "step": stats["step"],
            "count": f"{stats['count']:,}",
            "mean": f"{to_float(stats['mean']):,.4f}",
            "median": f"{to_float(stats['median']):,.4f}",
            "min": f"{to_float(stats['min']):,.4f}",
            "max": f"{to_float(stats['max']):,.4f}",
            "std": f"{to_float(stats['std']):,.4f}",
            "q01": f"{to_float(stats['q01']):,.4f}",
            "q05": f"{to_float(stats['q05']):,.4f}",
            "q10": f"{to_float(stats['q10']):,.4f}",
            "q25": f"{to_float(stats['q25']):,.4f}",
            "q75": f"{to_float(stats['q75']):,.4f}",
            "q90": f"{to_float(stats['q90']):,.4f}",
            "q95": f"{to_float(stats['q95']):,.4f}",
            "q99": f"{to_float(stats['q99']):,.4f}",
        }
    )



r"""Contain the implementation of a section to analyze the temporal
drift of a column with continuous values."""

from __future__ import annotations

__all__ = [
    "ColumnContinuousTemporalDriftSection",
    "create_section_template",
    "create_temporal_drift_figure",
]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot import hist_continuous2
from flamme.plot.utils import readable_xticklabels
from flamme.plot.utils.hist import adjust_nbins
from flamme.section import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.array import filter_range
from flamme.utils.figure import figure2html
from flamme.utils.mapping import sort_by_keys
from flamme.utils.range import find_range
from flamme.utils.temporal import to_temporal_frames

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

logger = logging.getLogger(__name__)


class ColumnContinuousTemporalDriftSection(BaseSection):
    r"""Implement a section that analyzes the temporal drift of a column
    with continuous values.

    Args:
        frame: The DataFrame with the data.
        column: The column name.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        nbins: The number of bins in the histogram.
        density: If True, draw and return a probability density:
            each bin will display the bin's raw count divided by the
            total number of counts and the bin width, so that the area
            under the histogram integrates to 1.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import numpy as np
    >>> import polars as pl
    >>> from flamme.section import ColumnContinuousTemporalDriftSection
    >>> rng = np.random.default_rng()
    >>> data = pl.DataFrame(
    ...     {
    ...         "col": rng.standard_normal(59),
    ...         "datetime": pl.datetime_range(
    ...             start=datetime(year=2018, month=1, day=1, tzinfo=timezone.utc),
    ...             end=datetime(year=2018, month=3, day=1, tzinfo=timezone.utc),
    ...             interval="1d",
    ...             closed="left",
    ...             eager=True,
    ...         ),
    ...     },
    ...     schema={
    ...         "col": pl.Float64,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> section = ColumnContinuousTemporalDriftSection(
    ...     frame=data, column="col", dt_column="date", period="1mo"
    ... )
    >>> section
    ColumnContinuousTemporalDriftSection(
      (column): col
      (dt_column): date
      (period): 1mo
      (nbins): None
      (density): False
      (yscale): auto
      (xmin): None
      (xmax): None
      (figsize): None
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        column: str,
        dt_column: str,
        period: str,
        nbins: int | None = None,
        density: bool = False,
        yscale: str = "auto",
        xmin: float | str | None = None,
        xmax: float | str | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._frame = frame
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._nbins = nbins
        self._density = density
        self._yscale = yscale
        self._xmin = xmin
        self._xmax = xmax
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "nbins": self._nbins,
                    "density": self._density,
                    "yscale": self._yscale,
                    "xmin": self._xmin,
                    "xmax": self._xmax,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def frame(self) -> pl.DataFrame:
        return self._frame

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def period(self) -> str:
        return self._period

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def density(self) -> bool:
        return self._density

    @property
    def xmin(self) -> float | str | None:
        return self._xmin

    @property
    def xmax(self) -> float | str | None:
        return self._xmax

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict[str, float]:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the temporal drift of {self._column}")
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "temporal_drift_figure": self._create_temporal_drift_figure(),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_temporal_drift_figure(self) -> str:
        fig = create_temporal_drift_figure(
            frame=self._frame,
            column=self._column,
            dt_column=self._dt_column,
            period=self._period,
            nbins=self._nbins,
            density=self._density,
            yscale=self._yscale,
            xmin=self._xmin,
            xmax=self._xmax,
            figsize=self._figsize,
        )
        return figure2html(fig, close_fig=True)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.continuous_drift import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal drift of continuous values for column <em>{{column}}</em>.

{{temporal_drift_figure}}

<p style="margin-top: 1rem;">
"""


def create_temporal_drift_figure(
    frame: pl.DataFrame,
    column: str,
    dt_column: str,
    period: str,
    nbins: int | None = None,
    density: bool = False,
    yscale: str = "linear",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Return the figure to analyze the temporal drift.

    Args:
        frame: The DataFrame with the data.
        column: The column name.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        nbins: The number of bins in the histogram.
        density: If True, draw and return a probability density:
            each bin will display the bin's raw count divided by the
            total number of counts and the bin width, so that the area
            under the histogram integrates to 1.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure or ``None`` if there is no valid data.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import numpy as np
    >>> import polars as pl
    >>> from flamme.section.continuous_drift import create_temporal_drift_figure
    >>> rng = np.random.default_rng()
    >>> data = pl.DataFrame(
    ...     {
    ...         "col": rng.standard_normal(59),
    ...         "datetime": pl.datetime_range(
    ...             start=datetime(year=2018, month=1, day=1, tzinfo=timezone.utc),
    ...             end=datetime(year=2018, month=3, day=1, tzinfo=timezone.utc),
    ...             interval="1d",
    ...             closed="left",
    ...             eager=True,
    ...         ),
    ...     },
    ...     schema={
    ...         "col": pl.Float64,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> fig = create_temporal_drift_figure(
    ...     frame=data, column="col", dt_column="date", period="1mo"
    ... )

    ```
    """
    if column not in frame or dt_column not in frame:
        return None
    array = frame[column].drop_nulls().drop_nans().to_numpy()
    if array.size == 0:
        return None

    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    nbins = adjust_nbins(nbins=nbins, array=filter_range(array, xmin=xmin, xmax=xmax))
    frames, steps = to_temporal_frames(
        frame=frame.select(column, dt_column), dt_column=dt_column, period=period
    )
    groups = sort_by_keys(dict(zip(steps, [fr[column].to_numpy() for fr in frames])))

    nrows = len(steps)
    steps1, steps2 = steps[:-1], steps[1:]
    if len(steps) == 2:
        nrows = 1
    if len(steps) > 2:
        steps1, steps2 = [steps[0], *steps1], [steps[-1], *steps2]
    if figsize is not None:
        figsize = (figsize[0], figsize[1] * nrows)
    fig, axes = plt.subplots(figsize=figsize, nrows=nrows)

    for i, (step1, step2) in enumerate(zip(steps1, steps2)):
        ax = axes[i] if nrows > 1 else axes
        hist_continuous2(
            ax=ax,
            array1=groups[step1],
            array2=groups[step2],
            label1=step1,
            label2=step2,
            xmin=xmin,
            xmax=xmax,
            nbins=nbins,
            density=density,
            yscale=yscale,
        )
        ax.set_title(f"{step1} vs {step2}")
        readable_xticklabels(ax, max_num_xticks=100)
    return fig



r"""Contain the implementation of a section to analyze a column with
continuous values."""

from __future__ import annotations

__all__ = [
    "ColumnContinuousSection",
    "create_section_template",
    "create_boxplot_figure",
    "create_histogram_figure",
    "create_stats_table",
    "to_array",
]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot import boxplot_continuous, hist_continuous
from flamme.plot.utils.hist import adjust_nbins
from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.array import filter_range, nonnan
from flamme.utils.figure import figure2html
from flamme.utils.range import find_range
from flamme.utils.stats import compute_statistics_continuous

if TYPE_CHECKING:
    from collections.abc import Sequence

    import numpy as np
    import polars as pl


logger = logging.getLogger(__name__)


class ColumnContinuousSection(BaseSection):
    r"""Implement a section that analyzes a continuous distribution of
    values.

    Args:
        series: The series/column to analyze.
        column: The column name.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'/'symlog'`` scale is chosen based
            on the distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.section import ColumnContinuousSection
    >>> section = ColumnContinuousSection(
    ...     series=pl.Series([None, *list(range(101)), None]), column="col"
    ... )
    >>> section
    ColumnContinuousSection(
      (column): col
      (nbins): None
      (yscale): auto
      (xmin): None
      (xmax): None
      (figsize): None
    )
    >>> section.get_statistics()
    {'count': 103, 'nunique': 102, 'num_non_nulls': 101, 'num_nulls': 2,
     'mean': 50.0, 'std': 29.15..., 'skewness': 0.0, 'kurtosis': -1.20..., 'min': 0.0,
     'q001': 0.1, 'q01': 1.0, 'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0,
     'q75': 75.0, 'q90': 90.0, 'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0,
     '>0': 100, '<0': 0, '=0': 1}

    ```
    """

    def __init__(
        self,
        series: pl.Series,
        column: str,
        nbins: int | None = None,
        yscale: str = "auto",
        xmin: float | str | None = None,
        xmax: float | str | None = None,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._series = series
        self._column = column
        self._nbins = nbins
        self._yscale = yscale
        self._xmin = xmin
        self._xmax = xmax
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "nbins": self._nbins,
                    "yscale": self._yscale,
                    "xmin": self._xmin,
                    "xmax": self._xmax,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def yscale(self) -> str:
        return self._yscale

    @property
    def nbins(self) -> int | None:
        return self._nbins

    @property
    def series(self) -> pl.Series:
        return self._series

    @property
    def xmin(self) -> float | str | None:
        return self._xmin

    @property
    def xmax(self) -> float | str | None:
        return self._xmax

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict[str, float]:
        return compute_statistics_continuous(self._series)

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(f"Rendering the continuous distribution of {self._column}")
        stats = self.get_statistics()
        null_values_pct = (
            f"{100 * stats['num_nulls'] / stats['count']:.2f}" if stats["count"] > 0 else "N/A"
        )
        xmin, xmax = find_range(to_array(self._series), xmin=self._xmin, xmax=self._xmax)
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "table": create_stats_table(stats=stats, column=self._column),
                "total_values": f"{stats['count']:,}",
                "unique_values": f"{stats['nunique']:,}",
                "null_values": f"{stats['num_nulls']:,}",
                "null_values_pct": null_values_pct,
                "histogram_figure": self._create_histogram_figure(),
                "boxplot_figure": self._create_boxplot_figure(),
                "min_value": f"{stats['min']:,}",
                "max_value": f"{stats['max']:,}",
                "xmin": f"{xmin:,}",
                "xmax": f"{xmax:,}",
                "dtype": str(self._series.dtype),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_boxplot_figure(self) -> str:
        fig = create_boxplot_figure(
            series=self._series,
            xmin=self._xmin,
            xmax=self._xmax,
            figsize=self._figsize,
        )
        return figure2html(fig, close_fig=True)

    def _create_histogram_figure(self) -> str:
        fig = create_histogram_figure(
            series=self._series,
            column=self._column,
            nbins=self._nbins,
            yscale=self._yscale,
            xmin=self._xmin,
            xmax=self._xmax,
            figsize=self._figsize,
        )
        return figure2html(fig, close_fig=True)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.continuous import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the distribution of continuous values for column <em>{{column}}</em>.

<ul>
  <li> <b>total values:</b> {{total_values}} </li>
  <li> <b>number of unique values:</b> {{unique_values}} </li>
  <li> <b>number of null values:</b> {{null_values}} / {{total_values}} ({{null_values_pct}}%) </li>
  <li> <b>range of values:</b> [{{min_value}}, {{max_value}}] </li>
  <li> <b>data type:</b> <em>{{dtype}}</em> </li>
</ul>

The histogram shows the distribution of values in the range [{{xmin}}, {{xmax}}].

{{histogram_figure}}
{{boxplot_figure}}
{{table}}

<p style="margin-top: 1rem;">
"""


def create_boxplot_figure(
    series: pl.Series,
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Return a boxplot figure.

    Args:
        series: The series/column to analyze.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.section.continuous import create_boxplot_figure
    >>> fig = create_boxplot_figure(series=pl.Series([None, *list(range(101)), None]))

    ```
    """
    array = to_array(series)
    if array.size == 0:
        return None
    if figsize is not None:
        figsize = (figsize[0], figsize[0] / 10)
    fig, ax = plt.subplots(figsize=figsize)
    boxplot_continuous(ax=ax, array=array, xmin=xmin, xmax=xmax)
    return fig


def create_histogram_figure(
    series: pl.Series,
    column: str,
    nbins: int | None = None,
    yscale: str = "linear",
    xmin: float | str | None = None,
    xmax: float | str | None = None,
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Return a histogram figure.

    Args:
        series: The series/column to analyze.
        column: The column name.
        nbins: The number of bins in the histogram.
        yscale: The y-axis scale. If ``'auto'``, the
            ``'linear'`` or ``'log'`` scale is chosen based on the
            distribution.
        xmin: The minimum value of the range or its
            associated quantile. ``q0.1`` means the 10% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        xmax: The maximum value of the range or its
            associated quantile. ``q0.9`` means the 90% quantile.
            ``0`` is the minimum value and ``1`` is the maximum value.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.section.continuous import create_histogram_figure
    >>> fig = create_histogram_figure(
    ...     series=pl.Series([None, *list(range(101)), None]), column="col"
    ... )

    ```
    """
    array = nonnan(to_array(series))
    if array.size == 0:
        return None
    fig, ax = plt.subplots(figsize=figsize)
    xmin, xmax = find_range(array, xmin=xmin, xmax=xmax)
    nbins = adjust_nbins(nbins=nbins, array=filter_range(array, xmin=xmin, xmax=xmax))
    hist_continuous(
        ax=ax,
        array=array,
        nbins=nbins,
        xmin=xmin,
        xmax=xmax,
        yscale=yscale,
    )
    ax.set_title(f"data distribution for column {column}")
    return fig


def create_stats_table(stats: dict, column: str) -> str:
    r"""Create the HTML code of the table with statistics.

    Args:
        stats: Specifies a dictionary with the statistics.
        column: The column name.

    Returns:
        The HTML code of the table.

    Example usage:

    ```pycon

    >>> from flamme.section.continuous import create_stats_table
    >>> table = create_stats_table(
    ...     column="col",
    ...     stats={
    ...         "count": 101,
    ...         "nunique": 101,
    ...         "num_non_nulls": 101,
    ...         "num_nulls": 0,
    ...         "mean": 50.0,
    ...         "std": 29.15,
    ...         "skewness": 0.0,
    ...         "kurtosis": -1.20,
    ...         "min": 0.0,
    ...         "q001": 0.1,
    ...         "q01": 1.0,
    ...         "q05": 5.0,
    ...         "q10": 10.0,
    ...         "q25": 25.0,
    ...         "median": 50.0,
    ...         "q75": 75.0,
    ...         "q90": 90.0,
    ...         "q95": 95.0,
    ...         "q99": 99.0,
    ...         "q999": 99.9,
    ...         "max": 100.0,
    ...         ">0": 100,
    ...         "<0": 0,
    ...         "=0": 1,
    ...     },
    ... )

    ```
    """
    return Template(
        """<details>
    <summary>[show statistics]</summary>

    <p>The following table shows some statistics about the distribution for column {{column}}.

    <table class="table table-hover table-responsive w-auto" >
        <thead class="thead table-group-divider">
            <tr><th>stat</th><th>value</th></tr>
        </thead>
        <tbody class="tbody table-group-divider">
            <tr><th>count</th><td {{num_style}}>{{count}}</td></tr>
            <tr><th>mean</th><td {{num_style}}>{{mean}}</td></tr>
            <tr><th>std</th><td {{num_style}}>{{std}}</td></tr>
            <tr><th>skewness</th><td {{num_style}}>{{skewness}}</td></tr>
            <tr><th>kurtosis</th><td {{num_style}}>{{kurtosis}}</td></tr>
            <tr><th>min</th><td {{num_style}}>{{min}}</td></tr>
            <tr><th>quantile 0.1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 1%</th><td {{num_style}}>{{q01}}</td></tr>
            <tr><th>quantile 5%</th><td {{num_style}}>{{q05}}</td></tr>
            <tr><th>quantile 10%</th><td {{num_style}}>{{q10}}</td></tr>
            <tr><th>quantile 25%</th><td {{num_style}}>{{q25}}</td></tr>
            <tr><th>median</th><td {{num_style}}>{{median}}</td></tr>
            <tr><th>quantile 75%</th><td {{num_style}}>{{q75}}</td></tr>
            <tr><th>quantile 90%</th><td {{num_style}}>{{q90}}</td></tr>
            <tr><th>quantile 95%</th><td {{num_style}}>{{q95}}</td></tr>
            <tr><th>quantile 99%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>quantile 99.9%</th><td {{num_style}}>{{q99}}</td></tr>
            <tr><th>max</th><td {{num_style}}>{{max}}</td></tr>
            <tr><th>number of zeros</th><td {{num_style}}>{{num_zeros}}</td></tr>
            <tr><th>number of positive values</th><td {{num_style}}>{{num_pos}}</td></tr>
            <tr><th>number of negative values</th><td {{num_style}}>{{num_neg}}</td></tr>
            <tr class="table-group-divider"></tr>
        </tbody>
    </table>
</details>
"""
    ).render(
        {
            "column": column,
            "num_style": 'style="text-align: right;"',
            "count": f"{stats['count']:,}",
            "mean": f"{stats['mean']:,.4f}",
            "std": f"{stats['std']:,.4f}",
            "skewness": f"{stats['skewness']:,.4f}",
            "kurtosis": f"{stats['kurtosis']:,.4f}",
            "min": f"{stats['min']:,.4f}",
            "q001": f"{stats['q001']:,.4f}",
            "q01": f"{stats['q01']:,.4f}",
            "q05": f"{stats['q05']:,.4f}",
            "q10": f"{stats['q10']:,.4f}",
            "q25": f"{stats['q25']:,.4f}",
            "median": f"{stats['median']:,.4f}",
            "q75": f"{stats['q75']:,.4f}",
            "q90": f"{stats['q90']:,.4f}",
            "q95": f"{stats['q95']:,.4f}",
            "q99": f"{stats['q99']:,.4f}",
            "q999": f"{stats['q999']:,.4f}",
            "max": f"{stats['max']:,.4f}",
            "num_pos": f"{stats['>0']:,}",
            "num_neg": f"{stats['<0']:,}",
            "num_zeros": f"{stats['=0']:,}",
        }
    )


def to_array(series: pl.Series) -> np.ndarray:
    r"""Convert a series to a numpy array.

    Args:
        series: The series to convert.

    Returns:
        The converted array.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.section.continuous import to_array
    >>> array = to_array(series=pl.Series([None, *list(range(5)), None]))
    >>> array
    array([0, 1, 2, 3, 4])

    ```
    """
    return series.drop_nulls().to_numpy().ravel()



r"""Contain plotting functionalities."""

from __future__ import annotations

__all__ = [
    "bar_discrete",
    "bar_discrete_temporal",
    "boxplot_continuous",
    "boxplot_continuous_temporal",
    "hist_continuous",
    "hist_continuous2",
    "plot_cdf",
    "plot_null_temporal",
]

from flamme.plot.cdf import plot_cdf
from flamme.plot.continuous import (
    boxplot_continuous,
    boxplot_continuous_temporal,
    hist_continuous,
    hist_continuous2,
)
from flamme.plot.discrete import bar_discrete, bar_discrete_temporal
from flamme.plot.null_temp import plot_null_temporal



r"""Contain the implementation of sections to analyze the number null
values for each column."""

from __future__ import annotations

__all__ = [
    "ColumnTemporalNullValueSection",
    "create_section_template",
    "create_table_section",
    "create_temporal_null_figure",
    "create_temporal_null_figures",
    "create_temporal_null_table",
    "create_temporal_null_table_row",
]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping, str_indent
from grizz.utils.imports import is_tqdm_available
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot import plot_null_temporal
from flamme.plot.utils import readable_xticklabels
from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.figure import MISSING_FIGURE_MESSAGE, figure2html
from flamme.utils.null import compute_temporal_null_count

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

if is_tqdm_available():
    from tqdm import tqdm
else:  # pragma: no cover
    from grizz.utils.noop import tqdm

logger = logging.getLogger(__name__)


class ColumnTemporalNullValueSection(BaseSection):
    r"""Implement a section to analyze the temporal distribution of null
    values for all columns.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze. A plot is generated
            for each column.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        ncols: The number of columns.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section import ColumnTemporalNullValueSection
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={
    ...         "int": pl.Int64,
    ...         "float": pl.Float64,
    ...         "str": pl.String,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> section = ColumnTemporalNullValueSection(
    ...     frame=frame, columns=["float", "int", "str"], dt_column="datetime", period="M"
    ... )
    >>> section
    ColumnTemporalNullValueSection(
      (columns): ('float', 'int', 'str')
      (dt_column): datetime
      (period): M
      (ncols): 2
      (figsize): (7, 5)
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        columns: Sequence[str],
        dt_column: str,
        period: str,
        ncols: int = 2,
        figsize: tuple[float, float] = (7, 5),
    ) -> None:
        self._frame = frame
        self._columns = tuple(columns)
        self._dt_column = dt_column
        self._period = period
        self._ncols = min(ncols, len(self._columns))
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "columns": self._columns,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "ncols": self._ncols,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def frame(self) -> pl.DataFrame:
        r"""The DataFrame to analyze."""
        return self._frame

    @property
    def columns(self) -> tuple[str, ...]:
        r"""The columns to analyze."""
        return self._columns

    @property
    def dt_column(self) -> str:
        r"""The datetime column."""
        return self._dt_column

    @property
    def period(self) -> str:
        r"""The temporal period used to analyze the data."""
        return self._period

    @property
    def ncols(self) -> int:
        r"""The number of columns to show the figures."""
        return self._ncols

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Rendering the temporal null value distribution of the following columns: "
            f"{self._columns}\ndatetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._dt_column,
                "figure": create_temporal_null_figure(
                    frame=self._frame,
                    columns=self._columns,
                    dt_column=self._dt_column,
                    period=self._period,
                    ncols=self._ncols,
                    figsize=self._figsize,
                ),
                "table": create_table_section(
                    frame=self._frame,
                    columns=self._columns,
                    dt_column=self._dt_column,
                    period=self._period,
                ),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.null_temp_col import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of null values.
The column <em>{{column}}</em> is used as the temporal column.

{{figure}}

<p style="margin-top: 1rem;">

{{table}}

<p style="margin-top: 1rem;">
"""


def create_temporal_null_figure(
    frame: pl.DataFrame,
    columns: Sequence[str],
    dt_column: str,
    period: str,
    ncols: int = 2,
    figsize: tuple[float, float] = (7, 5),
) -> str:
    r"""Create a HTML representation of a figure with the temporal null
    value distribution.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze. A plot is generated
            for each column.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        ncols: The number of columns.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Returns:
        The HTML representation of the figure.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.null_temp_col import create_temporal_null_figure
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={
    ...         "int": pl.Int64,
    ...         "float": pl.Float64,
    ...         "str": pl.String,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> figures = create_temporal_null_figure(
    ...     frame=frame, columns=["float", "int", "str"], dt_column="datetime", period="1mo"
    ... )

    ```
    """
    if frame.is_empty():
        return MISSING_FIGURE_MESSAGE
    figures = create_temporal_null_figures(
        frame=frame, columns=columns, dt_column=dt_column, period=period, figsize=figsize
    )
    figures = add_column_to_figure(columns=columns, figures=figures)
    return Template(
        """<div class="container-fluid text-center">
  <div class="row align-items-start">
    {{columns}}
  </div>
</div>
"""
    ).render({"columns": "\n".join(split_figures_by_column(figures=figures, ncols=ncols))})


def create_temporal_null_figures(
    frame: pl.DataFrame,
    columns: Sequence[str],
    dt_column: str,
    period: str,
    figsize: tuple[float, float] = (7, 5),
) -> list[str]:
    r"""Create a HTML representation of each figure with the temporal
    null value distribution.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze. A plot is generated
            for each column.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Returns:
        The HTML representations of the figures.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.null_temp_col import create_temporal_null_figures
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={
    ...         "int": pl.Int64,
    ...         "float": pl.Float64,
    ...         "str": pl.String,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> figures = create_temporal_null_figures(
    ...     frame=frame, columns=["float", "int", "str"], dt_column="datetime", period="1mo"
    ... )

    ```
    """
    if frame.is_empty():
        return []

    figures = []
    for column in tqdm(columns, desc="generating figures"):
        fig, ax = plt.subplots(figsize=figsize)
        ax.set_title(f"column: {column}")

        nulls, totals, labels = compute_temporal_null_count(
            frame=frame, columns=[column], dt_column=dt_column, period=period
        )
        plot_null_temporal(ax=ax, labels=labels, nulls=nulls, totals=totals)
        readable_xticklabels(ax, max_num_xticks=50)
        figures.append(figure2html(fig, close_fig=True))

    return figures


def add_column_to_figure(columns: Sequence[str], figures: Sequence[str]) -> list[str]:
    r"""Add the column name to the HTML representation of the figure.

    Args:
        columns: The column names.
        figures: The HTML representations of each figure.

    Returns:
        The uplated HTML representations of each figure.

    Raises:
        RuntimeError: if the number of column names is different from
            the number of figures.

    Example usage:

    ```pycon

    >>> from flamme.section.null_temp_col import add_column_to_figure
    >>> figures = add_column_to_figure(columns=["col1", "col2"], figures=["fig1", "fig2"])

    ```
    """
    if len(columns) != len(figures):
        msg = (
            f"The number of column names is different from the number of figures: "
            f"{len(columns):,} vs{len(figures):,}"
        )
        raise RuntimeError(msg)
    outputs = []
    for i, (col, figure) in enumerate(zip(columns, figures)):
        outputs.append(f'<div style="text-align:center">({i}) {col}\n{figure}</div>')
    return outputs


def split_figures_by_column(figures: Sequence[str], ncols: int) -> list[str]:
    r"""Split the figures into multiple columns.

    Args:
        figures: The HTML representations of each figure.
        ncols: The number of columns.

    Returns:
        The columns.

    Example usage:

    ```pycon

    >>> from flamme.section.null_temp_col import split_figures_by_column
    >>> cols = split_figures_by_column(figures=["fig1", "fig2", "fig3"], ncols=2)

    ```
    """
    cols = []
    for i in range(ncols):
        figs = str_indent("\n<hr>\n".join(figures[i::ncols]))
        cols.append(f'<div class="col">\n  {figs}\n</div>')
    return cols


def create_table_section(
    frame: pl.DataFrame, columns: Sequence[str], dt_column: str, period: str
) -> str:
    r"""Return a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        The HTML representation of the table.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.null_temp_col import create_table_section
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={
    ...         "int": pl.Int64,
    ...         "float": pl.Float64,
    ...         "str": pl.String,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> table = create_table_section(
    ...     frame=frame, columns=["float", "int", "str"], dt_column="datetime", period="1mo"
    ... )

    ```
    """
    if frame.is_empty():
        return ""
    tables = []
    for column in columns:
        table = create_temporal_null_table(
            frame=frame, column=column, dt_column=dt_column, period=period
        )
        tables.append(f'<p style="margin-top: 1rem;">\n\n{table}\n')
    return Template(
        """<details>
    <summary>[show statistics per temporal period]</summary>

    <p style="margin-top: 1rem;">
    The following table shows some statistics for each period of column {{column}}.

    {{tables}}
</details>
"""
    ).render({"tables": "\n".join(tables)})


def create_temporal_null_table(
    frame: pl.DataFrame, column: str, dt_column: str, period: str
) -> str:
    r"""Return a HTML representation of a table with the temporal
    distribution of null values.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.

    Returns:
        The HTML representation of the table.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.null_temp_col import create_temporal_null_table
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={
    ...         "int": pl.Int64,
    ...         "float": pl.Float64,
    ...         "str": pl.String,
    ...         "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...     },
    ... )
    >>> table = create_temporal_null_table(
    ...     frame=frame, column="int", dt_column="datetime", period="1mo"
    ... )

    ```
    """
    if frame.is_empty():
        return ""
    nulls, totals, labels = compute_temporal_null_count(
        frame=frame, columns=[column], dt_column=dt_column, period=period
    )
    rows = [
        create_temporal_null_table_row(label=label, null=null, total=total)
        for label, null, total in zip(labels, nulls, totals)
    ]
    return Template(
        """<table class="table table-hover table-responsive w-auto" >
    <thead class="thead table-group-divider">
        <tr>
            <th colspan="6" style="text-align: center">column: {{column}}</th>
        </tr>
        <tr>
            <th>period</th>
            <th>number of null values</th>
            <th>number of non-null values</th>
            <th>total number of values</th>
            <th>percentage of null values</th>
            <th>percentage of non-null values</th>
        </tr>
    </thead>
    <tbody class="tbody table-group-divider">
        {{rows}}
        <tr class="table-group-divider"></tr>
    </tbody>
</table>
"""
    ).render({"rows": "\n".join(rows), "column": column, "period": period})


def create_temporal_null_table_row(label: str, null: int, total: int) -> str:
    r"""Return the HTML code of a new table row.

    Args:
        label: The label of the row.
        null: The number of null values.
        total: The total number of values.

    Returns:
        The HTML code of a row.

    Example usage:

    ```pycon

    >>> from flamme.section.null_temp_col import create_temporal_null_table_row
    >>> row = create_temporal_null_table_row(label="col", null=5, total=42)

    ```
    """
    non_null = total - null
    return Template(
        """<tr>
    <th>{{label}}</th>
    <td {{num_style}}>{{null}}</td>
    <td {{num_style}}>{{non_null}}</td>
    <td {{num_style}}>{{total}}</td>
    <td {{num_style}}>{{nulls_pct}}</td>
    <td {{num_style}}>{{non_null_pct}}</td>
</tr>"""
    ).render(
        {
            "num_style": 'style="text-align: right;"',
            "label": label,
            "nulls": f"{null:,}",
            "non_null": f"{non_null:,}",
            "total": f"{total:,}",
            "nulls_pct": f"{100 * null / total:.2f}%",
            "non_null_pct": f"{100 * non_null / total:.2f}%",
        }
    )



r"""Implement some utility functions to manage optional dependencies."""

from __future__ import annotations

__all__ = [
    "check_clickhouse_connect",
    "check_colorlog",
    "check_markdown",
    "clickhouse_connect_available",
    "colorlog_available",
    "is_clickhouse_connect_available",
    "is_colorlog_available",
    "is_markdown_available",
    "markdown_available",
]

from typing import TYPE_CHECKING, Any

from coola.utils.imports import decorator_package_available, package_available

if TYPE_CHECKING:
    from collections.abc import Callable


##############################
#     clickhouse_connect     #
##############################


def is_clickhouse_connect_available() -> bool:
    r"""Indicate if the ``clickhouse_connect`` package is installed or
    not.

    Returns:
        ``True`` if ``clickhouse_connect`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import is_clickhouse_connect_available
    >>> is_clickhouse_connect_available()

    ```
    """
    return package_available("clickhouse_connect")


def check_clickhouse_connect() -> None:
    r"""Check if the ``clickhouse_connect`` package is installed.

    Raises:
        RuntimeError: if the ``clickhouse_connect`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import check_clickhouse_connect
    >>> check_clickhouse_connect()

    ```
    """
    if not is_clickhouse_connect_available():
        msg = (
            "`clickhouse_connect` package is required but not installed. "
            "You can install `clickhouse_connect` package with the command:\n\n"
            "pip install clickhouse-connect\n"
        )
        raise RuntimeError(msg)


def clickhouse_connect_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if
    ``clickhouse_connect`` package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``clickhouse_connect`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import clickhouse_connect_available
    >>> @clickhouse_connect_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_clickhouse_connect_available)


####################
#     colorlog     #
####################


def is_colorlog_available() -> bool:
    r"""Indicate if the ``colorlog`` package is installed or not.

    Returns:
        ``True`` if ``colorlog`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import is_colorlog_available
    >>> is_colorlog_available()

    ```
    """
    return package_available("colorlog")


def check_colorlog() -> None:
    r"""Check if the ``colorlog`` package is installed.

    Raises:
        RuntimeError: if the ``colorlog`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import check_colorlog
    >>> check_colorlog()

    ```
    """
    if not is_colorlog_available():
        msg = (
            "`colorlog` package is required but not installed. "
            "You can install `colorlog` package with the command:\n\n"
            "pip install colorlog\n"
        )
        raise RuntimeError(msg)


def colorlog_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``colorlog``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``colorlog`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import colorlog_available
    >>> @colorlog_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_colorlog_available)


####################
#     markdown     #
####################


def is_markdown_available() -> bool:
    r"""Indicate if the ``markdown`` package is installed or not.

    Returns:
        ``True`` if ``markdown`` is available otherwise
            ``False``.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import is_markdown_available
    >>> is_markdown_available()

    ```
    """
    return package_available("markdown")


def check_markdown() -> None:
    r"""Check if the ``markdown`` package is installed.

    Raises:
        RuntimeError: if the ``markdown`` package is not
            installed.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import check_markdown
    >>> check_markdown()

    ```
    """
    if not is_markdown_available():
        msg = (
            "`markdown` package is required but not installed. "
            "You can install `markdown` package with the command:\n\n"
            "pip install markdown\n"
        )
        raise RuntimeError(msg)


def markdown_available(fn: Callable[..., Any]) -> Callable[..., Any]:
    r"""Implement a decorator to execute a function only if ``markdown``
    package is installed.

    Args:
        fn: The function to execute.

    Returns:
        A wrapper around ``fn`` if ``markdown`` package is
            installed, otherwise ``None``.

    Example usage:

    ```pycon

    >>> from flamme.utils.imports import markdown_available
    >>> @markdown_available
    ... def my_function(n: int = 0) -> int:
    ...     return 42 + n
    ...
    >>> my_function()

    ```
    """
    return decorator_package_available(fn, is_markdown_available)



r"""Implement an analyzer that generates a section to analyze the data
types of each column."""

from __future__ import annotations

__all__ = ["DataTypeAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import DataTypeSection
from flamme.utils.dtype import frame_types

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class DataTypeAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to find all the value types in each column.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.analyzer import DataTypeAnalyzer
    >>> analyzer = DataTypeAnalyzer()
    >>> analyzer
    DataTypeAnalyzer()
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [42, 1, 0, 1],
    ...         "float": [1.2, 4.2, float("nan"), 2.2],
    ...         "str": ["A", "B", "C", "D"],
    ...     },
    ...     schema={"int": pl.Int64, "float": pl.Float64, "str": pl.String},
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    DataTypeSection(
      (dtypes): {'int': Int64, 'float': Float64, 'str': String}
      (types): {'int': {<class 'int'>}, 'float': {<class 'float'>}, 'str': {<class 'str'>}}
    )

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def analyze(self, frame: pl.DataFrame) -> DataTypeSection:
        logger.info("Analyzing the data types...")
        return DataTypeSection(dtypes=dict(frame.schema), types=frame_types(frame))



r"""Contain the implementation of a section to analyze the temporal
distribution of a column with discrete values."""

from __future__ import annotations

__all__ = ["ColumnTemporalDiscreteSection", "create_section_template", "create_temporal_figure"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping
from jinja2 import Template
from matplotlib import pyplot as plt

from flamme.plot import bar_discrete_temporal
from flamme.section.base import BaseSection
from flamme.section.utils import (
    GO_TO_TOP,
    render_html_toc,
    tags2id,
    tags2title,
    valid_h_tag,
)
from flamme.utils.count import compute_temporal_value_counts
from flamme.utils.figure import figure2html

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl

logger = logging.getLogger(__name__)


class ColumnTemporalDiscreteSection(BaseSection):
    r"""Implement a section that analyzes the temporal distribution of a
    column with discrete values.

    Args:
        frame: The DataFrame to analyze.
        column: The column of the DataFrame to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section import ColumnTemporalDiscreteSection
    >>> section = ColumnTemporalDiscreteSection(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col": [1, 42, None, 42],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Int64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )
    >>> section
    ColumnTemporalDiscreteSection(
      (column): col
      (dt_column): datetime
      (period): 1mo
      (figsize): None
    )
    >>> section.get_statistics()
    {}

    ```
    """

    def __init__(
        self,
        frame: pl.DataFrame,
        column: str,
        dt_column: str,
        period: str,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._frame = frame
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._figsize = figsize

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "column": self._column,
                    "dt_column": self._dt_column,
                    "period": self._period,
                    "figsize": self._figsize,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    @property
    def column(self) -> str:
        return self._column

    @property
    def dt_column(self) -> str:
        return self._dt_column

    @property
    def period(self) -> str:
        return self._period

    @property
    def figsize(self) -> tuple[float, float] | None:
        r"""The individual figure size in pixels.

        The first dimension is the width and the second is the height.
        """
        return self._figsize

    def get_statistics(self) -> dict:
        return {}

    def render_html_body(self, number: str = "", tags: Sequence[str] = (), depth: int = 0) -> str:
        logger.info(
            f"Analyzing the temporal discrete distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        return Template(create_section_template()).render(
            {
                "go_to_top": GO_TO_TOP,
                "id": tags2id(tags),
                "depth": valid_h_tag(depth + 1),
                "title": tags2title(tags),
                "section": number,
                "column": self._column,
                "dt_column": self._dt_column,
                "period": self._period,
                "figure_occurrence": self._create_temporal_figure(proportion=False),
                "figure_proportion": self._create_temporal_figure(proportion=True),
            }
        )

    def render_html_toc(
        self, number: str = "", tags: Sequence[str] = (), depth: int = 0, max_depth: int = 1
    ) -> str:
        return render_html_toc(number=number, tags=tags, depth=depth, max_depth=max_depth)

    def _create_temporal_figure(self, proportion: bool) -> str:
        fig = create_temporal_figure(
            frame=self._frame,
            column=self._column,
            dt_column=self._dt_column,
            period=self._period,
            proportion=proportion,
            figsize=self._figsize,
        )
        return figure2html(fig, close_fig=True)


def create_section_template() -> str:
    r"""Return the template of the section.

    Returns:
        The section template.

    Example usage:

    ```pycon

    >>> from flamme.section.discrete_temp import create_section_template
    >>> template = create_section_template()

    ```
    """
    return """<h{{depth}} id="{{id}}">{{section}} {{title}} </h{{depth}}>

{{go_to_top}}

<p style="margin-top: 1rem;">
This section analyzes the temporal distribution of column <em>{{column}}</em>
by using the column <em>{{dt_column}}</em>.
The first figure shows the number of occurrences for each value in each temporal window,
and the second figure shows the proportion for each value in each temporal window.

{{figure_occurrence}}
{{figure_proportion}}

{{table}}

<p style="margin-top: 1rem;">
"""


def create_temporal_figure(
    frame: pl.DataFrame,
    column: str,
    dt_column: str,
    period: str,
    proportion: bool = False,
    figsize: tuple[float, float] | None = None,
) -> plt.Figure | None:
    r"""Create a figure with the temporal value distribution.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.
        proportion: If ``True``, it plots the normalized number of
            occurrences for each step.
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Returns:
        The generated figure or None if the data is empty.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.section.discrete_temp import create_temporal_figure
    >>> fig = create_temporal_figure(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col": [1, 42, None, 42],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Int64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col",
    ...     dt_column="datetime",
    ...     period="1mo",
    ... )

    ```
    """
    if frame.is_empty() or column not in frame or dt_column not in frame:
        return None

    counts, steps, values = compute_temporal_value_counts(
        frame=frame,
        column=column,
        dt_column=dt_column,
        period=period,
        drop_nulls=True,
    )

    fig, ax = plt.subplots(figsize=figsize)
    bar_discrete_temporal(ax=ax, counts=counts, steps=steps, values=values, proportion=proportion)
    ax.set_title(f"Temporal distribution for column {column}")
    return fig



r"""Contain functionalities to plot the temporal distribution of the
number of missing values."""

from __future__ import annotations

__all__ = ["plot_null_temporal"]

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from collections.abc import Sequence

    from matplotlib.axes import Axes


def plot_null_temporal(ax: Axes, nulls: Sequence, totals: Sequence, labels: Sequence) -> None:
    r"""Plot the temporal distribution of the number of missing values.

    ``nulls``, ``totals``, and ``labels`` must have the same length
    and have the same order.

    Args:
        ax: The Axes object that encapsulates all the elements of an
            individual (sub-)plot in a figure.
        nulls: The number of null values for each temporal period.
        totals: The number of total values for each temporal period.
        labels: The labels for each temporal period.

    Raises:
        RuntimeError: if ``nulls``, ``totals``, and ``labels`` have
            different lengths.

    Example usage:

    ```pycon

    >>> from matplotlib import pyplot as plt
    >>> from flamme.plot import plot_null_temporal
    >>> fig, ax = plt.subplots()
    >>> plot_null_temporal(
    ...     ax, nulls=[1, 2, 3, 4], totals=[10, 12, 14, 16], labels=["jan", "feb", "mar", "apr"]
    ... )

    ```
    """
    if len(nulls) != len(totals):
        msg = f"nulls ({len(nulls):,}) and totals ({len(totals):,}) have different lengths"
        raise RuntimeError(msg)
    if len(labels) != len(totals):
        msg = f"nulls ({len(nulls):,}) and labels ({len(labels):,}) have different lengths"
        raise RuntimeError(msg)
    if len(nulls) == 0:
        return

    labels = list(map(str, labels))
    nulls = np.asarray(nulls)
    totals = np.asarray(totals)

    color = "tab:blue"
    x = np.arange(len(labels))
    ax.set_ylabel("number of null/total values", color=color)
    ax.tick_params(axis="y", labelcolor=color)
    ax.bar(x=x, height=totals, color="tab:cyan", alpha=0.5, label="total")
    ax.bar(x=x, height=nulls, color=color, alpha=0.8, label="null")
    ax.legend()

    ax2 = ax.twinx()
    color = "black"
    ax2.set_ylabel("percentage", color=color)
    ax2.tick_params(axis="y", labelcolor=color)
    ax2.plot(x, nulls / totals, "o-", color=color)

    ax.set_xticks(x, labels=labels)
    ax.set_xlim(-0.5, len(labels) - 0.5)



r"""Define some utility functions for testing."""

from __future__ import annotations

__all__ = ["clickhouse_connect_available", "colorlog_available", "markdown_available"]

import pytest

from flamme.utils.imports import (
    is_clickhouse_connect_available,
    is_colorlog_available,
    is_markdown_available,
)

clickhouse_connect_available = pytest.mark.skipif(
    not is_clickhouse_connect_available(), reason="requires clickhouse_connect"
)
colorlog_available = pytest.mark.skipif(not is_colorlog_available(), reason="requires colorlog")
markdown_available = pytest.mark.skipif(not is_markdown_available(), reason="requires markdown")



r"""Contain statistics utility functions."""

from __future__ import annotations

__all__ = [
    "compute_statistics_continuous",
    "compute_statistics_continuous_array",
    "compute_statistics_continuous_series",
    "quantile",
]

from typing import TYPE_CHECKING

import numpy as np
import polars as pl
from scipy.stats import kurtosis, skew

from flamme.utils.array import nonnan

if TYPE_CHECKING:
    from collections.abc import Sequence


def compute_statistics_continuous(data: np.ndarray | pl.Series) -> dict[str, float]:
    r"""Return several descriptive statistics for the data with
    continuous values.

    Args:
        data: The data to analyze.

    Returns:
        The descriptive statistics for the input data.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from flamme.utils.stats import compute_statistics_continuous
    >>> compute_statistics_continuous(np.arange(101))
    {'count': 101, 'nunique': 101, 'num_non_nulls': 101, 'num_nulls': 0,
     'mean': 50.0, 'std': 29.15...,
     'skewness': 0.0, 'kurtosis': -1.20..., 'min': 0.0, 'q001': 0.1, 'q01': 1.0,
     'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0, 'q75': 75.0, 'q90': 90.0,
     'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0, '>0': 100, '<0': 0, '=0': 1}

    ```
    """
    if isinstance(data, pl.Series):
        return compute_statistics_continuous_series(data)
    return compute_statistics_continuous_array(data)


def compute_statistics_continuous_array(array: np.ndarray) -> dict[str, float]:
    r"""Return several descriptive statistics for the data with
    continuous values.

    Args:
        array: The data to analyze.

    Returns:
        The descriptive statistics for the input data.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from flamme.utils.stats import compute_statistics_continuous_array
    >>> compute_statistics_continuous_array(np.arange(101))
    {'count': 101, 'nunique': 101, 'num_non_nulls': 101, 'num_nulls': 0,
     'mean': 50.0, 'std': 29.15...,
     'skewness': 0.0, 'kurtosis': -1.20..., 'min': 0.0, 'q001': 0.1, 'q01': 1.0,
     'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0, 'q75': 75.0, 'q90': 90.0,
     'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0, '>0': 100, '<0': 0, '=0': 1}

    ```
    """
    array = array.ravel().astype(np.float64)
    array_nonnan = nonnan(array)
    stats = {
        "count": int(array.size),
        "nunique": int(np.unique(array).size),
        "num_non_nulls": int(array_nonnan.size),
    }
    stats["num_nulls"] = stats["count"] - stats["num_non_nulls"]
    if array_nonnan.size == 0:
        return stats | {
            "mean": float("nan"),
            "std": float("nan"),
            "skewness": float("nan"),
            "kurtosis": float("nan"),
            "min": float("nan"),
            "q001": float("nan"),
            "q01": float("nan"),
            "q05": float("nan"),
            "q10": float("nan"),
            "q25": float("nan"),
            "median": float("nan"),
            "q75": float("nan"),
            "q90": float("nan"),
            "q95": float("nan"),
            "q99": float("nan"),
            "q999": float("nan"),
            "max": float("nan"),
            ">0": 0,
            "<0": 0,
            "=0": 0,
        }
    quantiles = quantile(
        array_nonnan, q=[0.001, 0.01, 0.05, 0.1, 0.25, 0.75, 0.9, 0.95, 0.99, 0.999]
    )
    return stats | {
        "mean": np.mean(array_nonnan).item(),
        "std": np.std(array_nonnan).item(),
        "skewness": float(skew(array_nonnan)),
        "kurtosis": float(kurtosis(array_nonnan)),
        "min": np.min(array_nonnan).item(),
        "q001": quantiles[0.001],
        "q01": quantiles[0.01],
        "q05": quantiles[0.05],
        "q10": quantiles[0.1],
        "q25": quantiles[0.25],
        "median": np.median(array_nonnan).item(),
        "q75": quantiles[0.75],
        "q90": quantiles[0.9],
        "q95": quantiles[0.95],
        "q99": quantiles[0.99],
        "q999": quantiles[0.999],
        "max": np.max(array_nonnan).item(),
        ">0": (array > 0).sum().item(),
        "<0": (array < 0).sum().item(),
        "=0": (array == 0).sum().item(),
    }


def compute_statistics_continuous_series(series: pl.Series) -> dict[str, float]:
    r"""Return several descriptive statistics for the data with
    continuous values.

    Args:
        series: The series to analyze.

    Returns:
        The descriptive statistics for the input data.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.utils.stats import compute_statistics_continuous_series
    >>> compute_statistics_continuous_series(pl.Series(list(range(101))))
    {'count': 101, 'nunique': 101, 'num_non_nulls': 101, 'num_nulls': 0,
     'mean': 50.0, 'std': 29.15...,
     'skewness': 0.0, 'kurtosis': -1.20..., 'min': 0.0, 'q001': 0.1, 'q01': 1.0,
     'q05': 5.0, 'q10': 10.0, 'q25': 25.0, 'median': 50.0, 'q75': 75.0, 'q90': 90.0,
     'q95': 95.0, 'q99': 99.0, 'q999': 99.9, 'max': 100.0, '>0': 100, '<0': 0, '=0': 1}

    ```
    """
    stats = {
        "count": int(series.shape[0]),
        "nunique": series.n_unique(),
        "num_nulls": int(series.null_count()),
    }
    stats["num_non_nulls"] = stats["count"] - stats["num_nulls"]
    return compute_statistics_continuous_array(series.drop_nulls().to_numpy()) | stats


def quantile(array: np.ndarray, q: Sequence[float]) -> dict[float, float]:
    r"""Compute the q-th quantile of the data.

    Args:
        array: The input data.
        q: The quantiles to compute. Values must be between 0 and 1
            inclusive.

    Returns:
        A dictionary with the quantiles values.

    Example usage:

    ```pycon

    >>> import numpy as np
    >>> from flamme.utils.stats import quantile
    >>> quantile(np.arange(101), q=[0.001, 0.01, 0.05, 0.1, 0.25, 0.75, 0.9, 0.95, 0.99, 0.999])
    {0.001: 0.1, 0.01: 1.0, 0.05: 5.0, 0.1: 10.0, 0.25: 25.0, 0.75: 75.0,
     0.9: 90.0, 0.95: 95.0, 0.99: 99.0, 0.999: 99.9}

    ```
    """
    array = array.ravel()
    if array.size == 0:
        return {v: float("nan") for v in q}
    return dict(zip(q, np.quantile(array.astype(np.float64), q).tolist()))



r"""Implement an analyzer that generates a section to analyze the number
of null values."""

from __future__ import annotations

__all__ = ["NullValueAnalyzer"]

import logging
from typing import TYPE_CHECKING

import numpy as np

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import NullValueSection
from flamme.utils.null import compute_null_count

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class NullValueAnalyzer(BaseAnalyzer):
    r"""Implement a null value analyzer.

    Args:
        figsize: The figure size in inches. The first
            dimension is the width and the second is the height.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.analyzer import NullValueAnalyzer
    >>> analyzer = NullValueAnalyzer()
    >>> analyzer
    NullValueAnalyzer(figsize=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "int": [None, 1, 0, 1],
    ...         "str": ["A", "B", None, None],
    ...     },
    ...     schema={"float": pl.Float64, "int": pl.Int64, "str": pl.String},
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    NullValueSection(
      (columns): ('float', 'int', 'str')
      (null_count): array([1, 1, 2])
      (total_count): array([4, 4, 4])
      (figsize): None
    )

    ```
    """

    def __init__(self, figsize: tuple[float, float] | None = None) -> None:
        self._figsize = figsize

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(figsize={self._figsize})"

    def analyze(self, frame: pl.DataFrame) -> NullValueSection:
        logger.info("Analyzing the null value distribution of all columns...")
        nrows, ncols = frame.shape
        return NullValueSection(
            columns=list(frame.columns),
            null_count=compute_null_count(frame),
            total_count=np.full((ncols,), nrows),
            figsize=self._figsize,
        )



r"""Contain utility functions to manage data types."""

from __future__ import annotations

__all__ = [
    "frame_types",
    "series_types",
]

import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import polars as pl


logger = logging.getLogger(__name__)


def frame_types(frame: pl.DataFrame) -> dict[str, set[type]]:
    r"""Return the value types per column.

    Args:
        frame: The DataFrame to analyze.

    Returns:
        A dictionary with the value types for each column.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.utils.dtype import frame_types
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [2, 1, 0, 1],
    ...         "float": [1.2, 4.2, float("nan"), 2.2],
    ...     },
    ...     schema={"int": pl.Int64, "float": pl.Float64},
    ... )
    >>> types = frame_types(frame)
    >>> types
    {'int': {<class 'int'>}, 'float': {<class 'float'>}}

    ```
    """
    return {col: series_types(frame[col]) for col in frame.columns}


def series_types(series: pl.Series) -> set[type]:
    r"""Return the value types in a ``polars.Series``.

    Args:
        series: The DataFrame to analyze.

    Returns:
        A dictionary with the value types for each column.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from flamme.utils.dtype import series_types
    >>> coltypes = series_types(pl.Series([1.2, 4.2, float("nan"), 2.2]))
    >>> coltypes
    {<class 'float'>}

    ```
    """
    return {type(x) for x in series.to_list()}


TYPE_NAMES = {}


def compact_type_name(typ: type) -> str:
    r"""Return a compact type name when possible.

    Args:
        typ: The input type.

    Returns:
        The compact type name.

    ```pycon

    >>> import polars as pl
    >>> from flamme.utils.dtype import compact_type_name
    >>> name = compact_type_name(int)
    >>> name
    int

    ```
    """
    name = str(typ).split("'", maxsplit=2)[1].rsplit("'", maxsplit=2)[0]
    return TYPE_NAMES.get(name, name)



r"""Implement an analyzer to analyze the temporal drift of a column with
continuous values."""

from __future__ import annotations

__all__ = ["ColumnTemporalDriftDiscreteAnalyzer"]

import logging
from typing import TYPE_CHECKING

from flamme.analyzer.base import BaseAnalyzer
from flamme.section import ColumnTemporalDriftDiscreteSection, EmptySection

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class ColumnTemporalDriftDiscreteAnalyzer(BaseAnalyzer):
    r"""Implement an analyzer to show the temporal distribution of
    discrete values.

    Args:
        column: The column to analyze.
        dt_column: The datetime column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or
            daily.
        proportion: If ``True``, it plots the normalized number of
            occurrences for each step.
        figsize: The figure size in inches. The first dimension
            is the width and the second is the height.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from flamme.analyzer import ColumnTemporalDriftDiscreteAnalyzer
    >>> analyzer = ColumnTemporalDriftDiscreteAnalyzer(
    ...     column="col", dt_column="datetime", period="1mo"
    ... )
    >>> analyzer
    ColumnTemporalDriftDiscreteAnalyzer(column=col, dt_column=datetime, period=1mo, proportion=False, figsize=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col": [1, 42, None, 42],
    ...         "datetime": [
    ...             datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...             datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...         ],
    ...     },
    ...     schema={"col": pl.Int64, "datetime": pl.Datetime(time_unit="us", time_zone="UTC")},
    ... )
    >>> section = analyzer.analyze(frame)
    >>> section
    ColumnTemporalDriftDiscreteSection(
      (column): col
      (dt_column): datetime
      (period): 1mo
      (proportion): False
      (figsize): None
    )

    ```
    """

    def __init__(
        self,
        column: str,
        dt_column: str,
        period: str,
        proportion: bool = False,
        figsize: tuple[float, float] | None = None,
    ) -> None:
        self._column = column
        self._dt_column = dt_column
        self._period = period
        self._proportion = proportion
        self._figsize = figsize

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(column={self._column}, "
            f"dt_column={self._dt_column}, period={self._period}, "
            f"proportion={self._proportion}, figsize={self._figsize})"
        )

    def analyze(self, frame: pl.DataFrame) -> ColumnTemporalDriftDiscreteSection | EmptySection:
        logger.info(
            f"Analyzing the temporal discrete distribution of {self._column} | "
            f"datetime column: {self._dt_column} | period: {self._period}"
        )
        for column in [self._column, self._dt_column]:
            if column not in frame:
                logger.info(
                    "Skipping temporal discrete distribution analysis because the column "
                    f"({column}) is not in the DataFrame"
                )
                return EmptySection()
        if self._column == self._dt_column:
            logger.info(
                "Skipping temporal discrete distribution analysis because the datetime column "
                f"({self._column}) is the column to analyze"
            )
            return EmptySection()
        return ColumnTemporalDriftDiscreteSection(
            column=self._column,
            frame=frame,
            dt_column=self._dt_column,
            period=self._period,
            proportion=self._proportion,
            figsize=self._figsize,
        )
